import abc
from typing import TypeVar
from typing_extensions import Self, override

import equinox as eqx
from jaxtyping import PRNGKeyArray

from .._pose import AbstractPose


T = TypeVar("T")


class AbstractVolumeParametrization(eqx.Module, strict=True):
    """Abstract interface for a parametrization of a volume. Specifically,
    the cryo-EM image formation process typically starts with a *scattering potential*.
    "Volumes" and "scattering potentials" in cryoJAX are synonymous.

    !!! info
        In, `cryojax`, potentials should be built in units of *inverse length squared*,
        $[L]^{-2}$. This rescaled potential is defined to be

        $$U(\\mathbf{r}) = \\frac{m_0 e}{2 \\pi \\hbar^2} V(\\mathbf{r}),$$

        where $V$ is the electrostatic potential energy, $\\mathbf{r}$ is a positional
        coordinate, $m_0$ is the electron rest mass, and $e$ is the electron charge.

        For a single atom, this rescaled potential has the advantage that under usual
        scattering approximations (i.e. the first-born approximation), the
        fourier transform of this quantity is closely related to tabulated electron scattering
        factors. In particular, for a single atom with scattering factor $f^{(e)}(\\mathbf{q})$
        and scattering vector $\\mathbf{q}$, its rescaled potential is equal to

        $$U(\\mathbf{r}) = \\mathcal{F}^{-1}[f^{(e)}(\\boldsymbol{\\xi} / 2)](\\mathbf{r}),$$

        where $\\boldsymbol{\\xi} = 2 \\mathbf{q}$ is the wave vector coordinate and
        $\\mathcal{F}^{-1}$ is the inverse fourier transform operator in the convention

        $$\\mathcal{F}[f](\\boldsymbol{\\xi}) = \\int d^3\\mathbf{r} \\ \\exp(2\\pi i \\boldsymbol{\\xi}\\cdot\\mathbf{r}) f(\\mathbf{r}).$$

        The rescaled potential $U$ gives the following time-independent schrodinger equation
        for the scattering problem,

        $$(\\nabla^2 + k^2) \\psi(\\mathbf{r}) = - 4 \\pi U(\\mathbf{r}) \\psi(\\mathbf{r}),$$

        where $k$ is the incident wavenumber of the electron beam.

        **References**:

        - For the definition of the rescaled potential, see
        Chapter 69, Page 2003, Equation 69.6 from *Hawkes, Peter W., and Erwin Kasper.
        Principles of Electron Optics, Volume 4: Advanced Wave Optics. Academic Press,
        2022.*
        - To work out the correspondence between the rescaled potential and the electron
        scattering factors, see the supplementary information from *Vulović, Miloš, et al.
        "Image formation modeling in cryo-electron microscopy." Journal of structural
        biology 183.1 (2013): 19-32.*
    """  # noqa: E501

    @abc.abstractmethod
    def get_representation(
        self, rng_key: PRNGKeyArray | None = None
    ) -> "AbstractVolumeRepresentation":
        """Core interface for computing the representation of
        the volume.

        **Arguments:**

        - `rng_key`:
            An optional RNG key for including noise / stochastic
            elements to volume simulation.
        """
        raise NotImplementedError


class AbstractVolumeRepresentation(AbstractVolumeParametrization, strict=True):
    """Abstract interface for the representation of a volume, such
    as atomic coordinates, voxels, or a neural network.

    Volume representations contain information of coordinates and may be
    passed to `AbstractVolumeIntegrator` classes for imaging.
    """

    @abc.abstractmethod
    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Rotate the coordinate system of the volume."""
        raise NotImplementedError

    @override
    def get_representation(self, rng_key: PRNGKeyArray | None = None) -> Self:
        """Since this class is itself an
        `AbstractVolumeRepresentation`, this function maps to the identity.

        **Arguments:**

        - `rng_key`:
            Not used in this implementation.
        """
        return self



from typing_extensions import override

from jaxtyping import Array, Complex, PRNGKeyArray

from ...jax_util import FloatLike
from .._image_config import AbstractImageConfig
from .._solvent_2d import AbstractRandomSolvent2D
from .._transfer_theory import ContrastTransferTheory
from .._volume import AbstractVolumeRepresentation
from .._volume_integrator import AbstractVolumeIntegrator
from .base_scattering_theory import AbstractScatteringTheory


class WeakPhaseScatteringTheory(AbstractScatteringTheory, strict=True):
    """Base linear image formation theory."""

    volume_integrator: AbstractVolumeIntegrator
    transfer_theory: ContrastTransferTheory
    solvent: AbstractRandomSolvent2D | None = None

    def __init__(
        self,
        volume_integrator: AbstractVolumeIntegrator,
        transfer_theory: ContrastTransferTheory,
        solvent: AbstractRandomSolvent2D | None = None,
    ):
        """**Arguments:**

        - `volume_integrator`: The method for integrating the scattering potential.
        - `transfer_theory`: The contrast transfer theory.
        - `solvent`: The model for the solvent.
        """
        self.volume_integrator = volume_integrator
        self.transfer_theory = transfer_theory
        self.solvent = solvent

    def compute_object_spectrum(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        # Compute the integrated potential
        fourier_in_plane_potential = self.volume_integrator.integrate(
            volume_representation, image_config, outputs_real_space=False
        )

        if rng_key is not None:
            # Get the potential of the specimen plus the ice
            if self.solvent is not None:
                fourier_in_plane_potential = self.solvent.compute_in_plane_potential(  # noqa: E501
                    rng_key,
                    fourier_in_plane_potential,
                    image_config,
                    input_is_rfft=self.volume_integrator.is_projection_approximation,
                )

        object_spectrum = image_config.interaction_constant * fourier_in_plane_potential

        return object_spectrum

    @override
    def compute_contrast_spectrum(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
        defocus_offset: FloatLike | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        object_spectrum = self.compute_object_spectrum(
            volume_representation, image_config, rng_key
        )
        contrast_spectrum = self.transfer_theory.propagate_object(  # noqa: E501
            object_spectrum,
            image_config,
            is_projection_approximation=self.volume_integrator.is_projection_approximation,
            defocus_offset=defocus_offset,
        )

        return contrast_spectrum

    @override
    def compute_intensity_spectrum(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
        defocus_offset: FloatLike | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        """Compute the squared wavefunction at the detector plane, given the
        contrast.
        """
        N1, N2 = image_config.padded_shape
        # ... compute the squared wavefunction directly from the image contrast
        # as |psi|^2 = 1 + 2C.
        contrast_spectrum = self.compute_contrast_spectrum(
            volume_representation, image_config, rng_key, defocus_offset=defocus_offset
        )
        intensity_spectrum = (2 * contrast_spectrum).at[0, 0].add(1.0 * N1 * N2)
        return intensity_spectrum



from typing_extensions import override

from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ...jax_util import FloatLike, error_if_not_fractional
from .._image_config import AbstractImageConfig
from .._multislice_integrator import AbstractMultisliceIntegrator
from .._transfer_theory import WaveTransferTheory
from .._volume import AbstractVolumeRepresentation
from .base_scattering_theory import AbstractWaveScatteringTheory


class MultisliceScatteringTheory(AbstractWaveScatteringTheory, strict=True):
    """A scattering theory using the multislice method."""

    volume_integrator: AbstractMultisliceIntegrator
    transfer_theory: WaveTransferTheory
    amplitude_contrast_ratio: Float[Array, ""]

    def __init__(
        self,
        volume_integrator: AbstractMultisliceIntegrator,
        transfer_theory: WaveTransferTheory,
        amplitude_contrast_ratio: FloatLike = 0.1,
    ):
        """**Arguments:**

        - `volume_integrator`: The multislice method.
        - `transfer_theory`: The wave transfer theory.
        - `amplitude_contrast_ratio`: The amplitude contrast ratio.
        """
        self.volume_integrator = volume_integrator
        self.transfer_theory = transfer_theory
        self.amplitude_contrast_ratio = error_if_not_fractional(amplitude_contrast_ratio)

    @override
    def compute_exit_wave(
        self,
        volume_representation: AbstractVolumeRepresentation,
        image_config: AbstractImageConfig,
        rng_key: PRNGKeyArray | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]:
        # Compute the wavefunction in the exit plane
        wavefunction = self.volume_integrator.integrate(
            volume_representation, image_config, self.amplitude_contrast_ratio
        )

        return wavefunction



"""
Base representations of volumes.
"""

import abc
from typing import TypeVar
from typing_extensions import Self

from jaxtyping import Float

from ....jax_util import NDArrayLike
from ..._pose import AbstractPose
from ..base_volume import AbstractVolumeRepresentation


T = TypeVar("T")


class AbstractAtomVolume(AbstractVolumeRepresentation, strict=True):
    """Abstract interface for a volume represented as a point-cloud."""

    @abc.abstractmethod
    def translate_to_pose(self, pose: AbstractPose) -> Self:
        raise NotImplementedError


class AbstractVoxelVolume(AbstractVolumeRepresentation, strict=True):
    """Abstract interface for a volume represented with voxels.

    !!! info

        If you are using a `volume` in a voxel representation
        pass, the voxel size *must* be passed as the
        `pixel_size` argument, e.g.

        ```python
        import cryojax.simulator as cxs
        from cryojax.io import read_array_from_mrc

        real_voxel_grid, voxel_size = read_array_from_mrc("example.mrc")
        volume = cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxel_grid)
        ...
        config = cxs.BasicImageConfig(shape, pixel_size=voxel_size, ...)
        ```

        If this is not done, the resulting
        image will be incorrect and *not* rescaled to the specified
        to the different pixel size.
    """

    @property
    @abc.abstractmethod
    def shape(self) -> tuple[int, ...]:
        """The shape of the voxel array."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_real_voxel_grid(
        cls, real_voxel_grid: Float[NDArrayLike, "dim dim dim"]
    ) -> Self:
        """Load an `AbstractVoxelStructure` from a 3D grid in
        real-space.
        """
        raise NotImplementedError



from typing import Any, Literal
from typing_extensions import override

import jax
import jax.numpy as jnp
from jaxtyping import Array, Float

from ...coordinates import make_frequency_grid
from ...jax_util import FloatLike, error_if_not_positive
from ...ndimage import convert_fftn_to_rfftn, ifftn, operators as op
from .._volume import IndependentAtomVolume
from .base_rendering import AbstractVolumeRenderFn


try:
    import jax_finufft as jnufft

    JAX_FINUFFT_IMPORT_ERROR = None
except ModuleNotFoundError as err:
    jnufft = None
    JAX_FINUFFT_IMPORT_ERROR = err


class FFTAtomRenderFn(AbstractVolumeRenderFn[IndependentAtomVolume], strict=True):
    """Render a voxel grid using non-uniform FFTs and convoluton."""

    shape: tuple[int, int, int]
    voxel_size: Float[Array, ""]
    frequency_grid: Float[Array, "_ _ _ 3"] | None
    sampling_mode: Literal["average", "point"]
    eps: float
    opts: Any

    def __init__(
        self,
        shape: tuple[int, int, int],
        voxel_size: FloatLike,
        *,
        frequency_grid: Float[Array, "_ _ _ 3"] | None = None,
        sampling_mode: Literal["average", "point"] = "average",
        eps: float = 1e-6,
        opts: Any = None,
    ):
        """**Arguments:**

        - `shape`:
            The shape of the resulting voxel grid.
        - `voxel_size`:
            The voxel size of the resulting voxel grid.
        - `frequency_grid`:
            An optional frequency grid for rendering the
            volume. If `None`, compute on the fly. The grid
            should be in inverse angstroms and have the zero
            frequency component in the center, i.e.

            ```python
            frequency_grid = jnp.fft.fftshift(
                make_frequency_grid(shape, voxel_size, outputs_rfft=False),
                axes=(0, 1, 2),
            )
            ```

        - `sampling_mode`:
            If `'average'`, convolve with a box function to sample the
            projected volume at a pixel to be the average value of the
            underlying continuous function. If `'point'`, the volume at
            a pixel will be point sampled.
        - `eps`:
            See [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
            for documentation.
        - `opts`:
            A `jax_finufft.options.Opts` or `jax_finufft.options.NestedOpts`
            dataclass.
            See [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
            for documentation.
        """
        self.shape = shape
        self.voxel_size = error_if_not_positive(jnp.asarray(voxel_size, dtype=float))
        self.frequency_grid = frequency_grid
        self.sampling_mode = sampling_mode
        self.eps = eps
        self.opts = opts

    @override
    def __call__(
        self,
        volume_representation: IndependentAtomVolume,
        *,
        outputs_real_space: bool = True,
        outputs_rfft: bool = False,
        fftshifted: bool = False,
    ) -> Float[Array, "{self.shape[0]} {self.shape[1]} {self.shape[2]}"]:
        """**Arguments:**

        - `volume_representation`:
            The `GaussianMixtureVolume`.
        - `outputs_real_space`:
            If `True`, return a voxel grid in real-space.
        - `outputs_rfft`:
            If `True`, return a fourier-space voxel grid transformed with
            `cryojax.ndimage.rfftn`. Otherwise, use `fftn`. Does nothing
            if `outputs_real_space = True`.
        - `fftshifted`:
            If `True`, return a fourier-space voxel grid with the zero
            frequency component in the center of the grid via
            `jax.numpy.fft.fftshift`. Otherwise, the zero frequency
            component is in the corner. Does nothing if
            `outputs_real_space = True`.
        """
        if self.frequency_grid is None:
            frequency_grid = jnp.fft.fftshift(
                make_frequency_grid(self.shape, self.voxel_size, outputs_rfftfreqs=False),
                axes=(0, 1, 2),
            )
        else:
            frequency_grid = self.frequency_grid
        proj_kernel = lambda pos, kernel: _render_with_nufft(
            self.shape,
            self.voxel_size,
            pos,
            kernel,
            frequency_grid,
            eps=self.eps,
            opts=self.opts,
        )
        # Compute projection over atom types
        fourier_voxel_grid = jax.tree.reduce(
            lambda x, y: x + y,
            jax.tree.map(
                proj_kernel,
                volume_representation.position_pytree,
                volume_representation.scattering_factor_pytree,
                is_leaf=lambda x: isinstance(x, op.AbstractFourierOperator),
            ),
        )
        if self.sampling_mode == "average":
            antialias_fn = op.FourierSinc(box_width=self.voxel_size)
            fourier_voxel_grid *= antialias_fn(frequency_grid)

        if outputs_real_space:
            return ifftn(jnp.fft.ifftshift(fourier_voxel_grid)).real
        else:
            if outputs_rfft:
                fourier_voxel_grid = convert_fftn_to_rfftn(
                    jnp.fft.ifftshift(fourier_voxel_grid), mode="real"
                )
                if fftshifted:
                    return jnp.fft.fftshift(fourier_voxel_grid, axes=(0, 1))
                else:
                    return fourier_voxel_grid
            else:
                if fftshifted:
                    return fourier_voxel_grid
                else:
                    return jnp.fft.ifftshift(fourier_voxel_grid)


def _render_with_nufft(shape, ps, pos, kernel, freqs, eps=1e-6, opts=None):
    assert jnufft is not None
    # Get x and y coordinates
    # Normalize coordinates betweeen -pi and pi
    nz, ny, nx = shape
    box_xyz = ps * jnp.asarray((nx, ny, nz))
    pos_periodic = 2 * jnp.pi * pos / box_xyz
    # Unpack
    x, y, z = pos_periodic[:, 0], pos_periodic[:, 1], pos_periodic[:, 2]
    n_atoms = x.size
    volume_element = ps**3
    # Compute
    fourier_projection = kernel(freqs) * (
        jnufft.nufft1(
            shape,
            jnp.full((n_atoms,), 1.0 + 0.0j),
            z,
            y,
            x,
            eps=eps,
            opts=opts,
            iflag=-1,
        )
        / volume_element
    )

    return fourier_projection



"""
Voxel-based representations of a volume.
"""

from typing import cast
from typing_extensions import Self, override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float

from ....coordinates import make_coordinate_grid, make_frequency_slice
from ....jax_util import NDArrayLike
from ....ndimage import (
    compute_spline_coefficients,
    crop_to_shape,
    fftn,
    pad_to_shape,
)
from ....ndimage.transforms import AbstractFilter
from ..._pose import AbstractPose
from .base_representations import AbstractVoxelVolume


class AbstractFourierVoxelVolume(AbstractVoxelVolume, strict=True):
    """Abstract interface for a voxel-based volume."""

    frequency_slice_in_pixels: eqx.AbstractVar[Float[Array, "1 dim dim 3"]]

    @override
    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Return a new volume with a rotated `frequency_slice_in_pixels`."""
        return eqx.tree_at(
            lambda d: d.frequency_slice_in_pixels,
            self,
            pose.rotate_coordinates(self.frequency_slice_in_pixels, inverse=inverse),
        )


class FourierVoxelGridVolume(AbstractFourierVoxelVolume, strict=True):
    """A 3D voxel grid in fourier-space."""

    fourier_voxel_grid: Complex[Array, "dim dim dim"]
    frequency_slice_in_pixels: Float[Array, "1 dim dim 3"]

    def __init__(
        self,
        fourier_voxel_grid: Complex[NDArrayLike, "dim dim dim"],
        frequency_slice_in_pixels: Float[NDArrayLike, "1 dim dim 3"],
    ):
        """**Arguments:**

        - `fourier_voxel_grid`:
            The cubic voxel grid in fourier space.
        - `frequency_slice_in_pixels`:
            The frequency slice coordinate system.
        """
        self.fourier_voxel_grid = jnp.asarray(fourier_voxel_grid, dtype=complex)
        self.frequency_slice_in_pixels = jnp.asarray(
            frequency_slice_in_pixels, dtype=float
        )

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the `fourier_voxel_grid`."""
        return cast(tuple[int, int, int], self.fourier_voxel_grid.shape)

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[NDArrayLike, "dim dim dim"],
        *,
        pad_scale: float = 1.0,
        pad_mode: str = "constant",
        filter: AbstractFilter | None = None,
    ) -> Self:
        """Load from a real-valued 3D voxel grid.

        **Arguments:**

        - `real_voxel_grid`: A voxel grid in real space.
        - `pad_scale`: Scale factor at which to pad `real_voxel_grid` before fourier
                     transform. Must be a value greater than `1.0`.
        - `pad_mode`: Padding method. See `jax.numpy.pad` for documentation.
        - `filter`: A filter to apply to the result of the fourier transform of
                  `real_voxel_grid`, i.e. `fftn(real_voxel_grid)`. Note that the zero
                  frequency component is assumed to be in the corner.
        """
        # Cast to jax array
        real_voxel_grid = jnp.asarray(real_voxel_grid, dtype=float)
        # Pad template
        if pad_scale < 1.0:
            raise ValueError("`pad_scale` must be greater than 1.0")
        # ... always pad to even size to avoid interpolation issues in
        # fourier slice extraction.
        padded_shape = cast(
            tuple[int, int, int],
            tuple([int(s * pad_scale) for s in real_voxel_grid.shape]),
        )
        padded_real_voxel_grid = pad_to_shape(
            real_voxel_grid, padded_shape, mode=pad_mode
        )
        # Load grid and coordinates. For now, do not store the
        # fourier grid only on the half space. Fourier slice extraction
        # does not currently work if rfftn is used.
        fourier_voxel_grid_with_zero_in_corner = (
            fftn(padded_real_voxel_grid)
            if filter is None
            else filter(fftn(padded_real_voxel_grid))
        )
        # ... store the grid with the zero frequency component in the center
        fourier_voxel_grid = jnp.fft.fftshift(fourier_voxel_grid_with_zero_in_corner)
        # ... create in-plane frequency slice on the half space
        frequency_slice = make_frequency_slice(
            cast(tuple[int, int], padded_real_voxel_grid.shape[:-1]),
            outputs_rfftfreqs=False,
        )

        return cls(fourier_voxel_grid, frequency_slice)


class FourierVoxelSplineVolume(AbstractFourierVoxelVolume, strict=True):
    """A 3D voxel grid in fourier-space, represented
    by spline coefficients.
    """

    spline_coefficients: Complex[Array, "coeff_dim coeff_dim coeff_dim"]
    frequency_slice_in_pixels: Float[Array, "1 dim dim 3"]

    def __init__(
        self,
        spline_coefficients: Complex[NDArrayLike, "coeff_dim coeff_dim coeff_dim"],
        frequency_slice_in_pixels: Float[NDArrayLike, "1 dim dim 3"],
    ):
        """**Arguments:**

        - `spline_coefficients`:
            The spline coefficents computed from the cubic voxel grid
            in fourier space. See `cryojax.ndimage.compute_spline_coefficients`.
        - `frequency_slice_in_pixels`:
            Frequency slice coordinate system.
            See `cryojax.coordinates.make_frequency_slice`.
        """
        self.spline_coefficients = jnp.asarray(spline_coefficients, dtype=complex)
        self.frequency_slice_in_pixels = jnp.asarray(
            frequency_slice_in_pixels, dtype=float
        )

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the original `fourier_voxel_grid` from which
        `coefficients` were computed.
        """
        return cast(
            tuple[int, int, int], tuple([s - 2 for s in self.spline_coefficients.shape])
        )

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[NDArrayLike, "dim dim dim"],
        *,
        pad_scale: float = 1.0,
        pad_mode: str = "constant",
        filter: AbstractFilter | None = None,
    ) -> Self:
        """Load from a real-valued 3D voxel grid.

        **Arguments:**

        - `real_voxel_grid`: A voxel grid in real space.
        - `pad_scale`: Scale factor at which to pad `real_voxel_grid` before fourier
                     transform. Must be a value greater than `1.0`.
        - `pad_mode`: Padding method. See `jax.numpy.pad` for documentation.
        - `filter`: A filter to apply to the result of the fourier transform of
                  `real_voxel_grid`, i.e. `fftn(real_voxel_grid)`. Note that the zero
                  frequency component is assumed to be in the corner.
        """
        # Cast to jax array
        real_voxel_grid = jnp.asarray(real_voxel_grid, dtype=float)
        # Pad template
        if pad_scale < 1.0:
            raise ValueError("`pad_scale` must be greater than 1.0")
        # ... always pad to even size to avoid interpolation issues in
        # fourier slice extraction.
        padded_shape = cast(
            tuple[int, int, int],
            tuple([int(s * pad_scale) for s in real_voxel_grid.shape]),
        )
        padded_real_voxel_grid = pad_to_shape(
            real_voxel_grid, padded_shape, mode=pad_mode
        )
        # Load grid and coordinates. For now, do not store the
        # fourier grid only on the half space. Fourier slice extraction
        # does not currently work if rfftn is used.
        fourier_voxel_grid_with_zero_in_corner = (
            fftn(padded_real_voxel_grid)
            if filter is None
            else filter(fftn(padded_real_voxel_grid))
        )
        # ... store the grid with the zero frequency component in the center
        fourier_voxel_grid = jnp.fft.fftshift(fourier_voxel_grid_with_zero_in_corner)
        # ... compute spline coefficients
        spline_coefficients = compute_spline_coefficients(fourier_voxel_grid)
        # ... create in-plane frequency slice on the half space
        frequency_slice = make_frequency_slice(
            cast(tuple[int, int], padded_real_voxel_grid.shape[:-1]),
            outputs_rfftfreqs=False,
        )

        return cls(spline_coefficients, frequency_slice)


class AbstractRealVoxelVolume(AbstractVoxelVolume, strict=True):
    """Abstract interface for a voxel-based volume."""

    coordinate_grid_in_pixels: eqx.AbstractVar[Float[Array, "dim dim dim 3"]]

    @override
    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Return a new volume with a rotated
        `coordinate_grid_in_pixels`.
        """
        return eqx.tree_at(
            lambda d: d.coordinate_grid_in_pixels,
            self,
            pose.rotate_coordinates(self.coordinate_grid_in_pixels, inverse=inverse),
        )


class RealVoxelGridVolume(AbstractRealVoxelVolume, strict=True):
    """A 3D voxel grid in real-space."""

    real_voxel_grid: Float[Array, "dim dim dim"]
    coordinate_grid_in_pixels: Float[Array, "dim dim dim 3"]

    def __init__(
        self,
        real_voxel_grid: Float[NDArrayLike, "dim dim dim"],
        coordinate_grid_in_pixels: Float[NDArrayLike, "dim dim dim 3"],
    ):
        """**Arguments:**

        - `real_voxel_grid`: The voxel grid in fourier space.
        - `coordinate_grid_in_pixels`: A coordinate grid.
        """
        self.real_voxel_grid = jnp.asarray(real_voxel_grid, dtype=float)
        self.coordinate_grid_in_pixels = jnp.asarray(
            coordinate_grid_in_pixels, dtype=float
        )

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the `real_voxel_grid`."""
        return cast(tuple[int, int, int], self.real_voxel_grid.shape)

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[NDArrayLike, "dim dim dim"],
        *,
        coordinate_grid_in_pixels: Float[Array, "dim dim dim 3"] | None = None,
        crop_scale: float | None = None,
    ) -> Self:
        """Load a `RealVoxelGridVolume` from a real-valued 3D voxel grid.

        **Arguments:**

        - `real_voxel_grid`: A voxel grid in real space.
        - `crop_scale`: Scale factor at which to crop `real_voxel_grid`.
                        Must be a value greater than `1`.
        """
        # Cast to jax array
        real_voxel_grid = jnp.asarray(real_voxel_grid, dtype=float)
        # Make coordinates if not given
        if coordinate_grid_in_pixels is None:
            # Option for cropping template
            if crop_scale is not None:
                if crop_scale < 1.0:
                    raise ValueError("`crop_scale` must be greater than 1.0")
                cropped_shape = cast(
                    tuple[int, int, int],
                    tuple([int(s / crop_scale) for s in real_voxel_grid.shape[-3:]]),
                )
                real_voxel_grid = crop_to_shape(real_voxel_grid, cropped_shape)
            coordinate_grid_in_pixels = make_coordinate_grid(real_voxel_grid.shape[-3:])

        return cls(real_voxel_grid, coordinate_grid_in_pixels)



from ._base_operator import (
    AbstractImageOperator as AbstractImageOperator,
    Constant as Constant,
    CustomOperator as CustomOperator,
    DiffImageOperator as DiffImageOperator,
    Empirical as Empirical,
    ProductImageOperator as ProductImageOperator,
    SumImageOperator as SumImageOperator,
)
from ._fourier_operator import (
    AbstractFourierOperator as AbstractFourierOperator,
    FourierExp2D as FourierExp2D,
    FourierGaussian as FourierGaussian,
    FourierGaussianWithRadialOffset as FourierGaussianWithRadialOffset,
    FourierOperatorLike as FourierOperatorLike,
    FourierSinc as FourierSinc,
    Lorenzian as Lorenzian,
    ZeroMode as ZeroMode,
)
from ._real_operator import (
    AbstractRealOperator as AbstractRealOperator,
    Gaussian2D as Gaussian2D,
    RealOperatorLike as RealOperatorLike,
)



"""
Routines for loading atomic structures.
Large amounts of the code are adapted from the ioSPI package
"""

import importlib.resources as pkg_resources
import os

import equinox as eqx
import jax.numpy as jnp
import numpy as np
import xarray as xr
from jaxtyping import Float, Int


class PengScatteringFactorParameters(eqx.Module, strict=True):
    """A convenience wrapper for instantiating the
    scattering factor parameters from Peng et al. (1996).

    To access scattering factors $a_i$ and $b_i$ given in
    the citation,

    ```python
    from cryojax.io import read_atoms_from_pdb
    from cryojax.constants import PengScatteringFactorParameters

    # Load positions of atoms and one-hot encoded atom names
    atom_positions, atom_types = read_atoms_from_pdb(...)
    parameters = PengScatteringFactorParameters(atom_types)
    print(parameters.a, parameters.b)  # a_i and b_i
    ```
    """

    a: Float[np.ndarray, " n_atoms 5"]
    b: Float[np.ndarray, " n_atoms 5"]

    def __init__(self, atomic_numbers: Int[np.ndarray, " n_atoms"]):
        """**Arguments:**

        - `atomic_numbers`:
            The atom types as an integer array.
        """
        scattering_factor_parameter_table = read_peng_scattering_factor_parameter_table()
        parameter_dict = extract_scattering_factor_parameters(
            atomic_numbers, scattering_factor_parameter_table
        )
        self.a = parameter_dict["a"]
        self.b = parameter_dict["b"]


def read_peng_scattering_factor_parameter_table() -> xr.Dataset:
    r"""Function to load the atomic scattering factor parameter
    table from "Robust Parameterization of Elastic and Absorptive
    Electron Atomic Scattering Factors" by Peng et al. (1996).

    **Returns:**

    The parameter table for parameters $\{a_i\}_{i = 1}^5$ and $\{b_i\}_{i = 1}^5$
    for each atom, described in the above reference. This is stored as an
    `xarray.Dataset`.
    """
    with pkg_resources.as_file(
        pkg_resources.files("cryojax").joinpath("constants")
    ) as path:
        scattering_factor_parameters = jnp.load(
            os.path.join(path, "peng1996_element_params.npy")
        )
    dims = ("n_elements", "n_gaussians")
    scattering_factor_parameter_table = xr.Dataset(
        data_vars=dict(
            a=xr.DataArray(scattering_factor_parameters[0], dims=dims),
            b=xr.DataArray(scattering_factor_parameters[1], dims=dims),
        ),
        attrs=dict(description="Scattering factor parameters from Peng et al. (1996)"),
    )

    return scattering_factor_parameter_table


def extract_scattering_factor_parameters(
    atom_types: Int[np.ndarray, " n_atoms"],
    scattering_factor_parameter_table: xr.Dataset | None = None,
) -> dict[str, Float[np.ndarray, " n_atoms n_scattering_factors"]]:
    """Gets the parameters for the scattering factor for each atom in
    `atom_types`.

    **Arguments:**

    - `atom_identitites`:
        Array containing the index of the one-hot encoded atom names.
        By default, Hydrogen is "1", Carbon is "6", Nitrogen is "7", etc.
    - `scattering_factor_parameter_table`:
        The table of scattering factors as an `xarray.Dataset`. By default, this
        is the tabulation from "Robust Parameterization of Elastic and
        Absorptive Electron Atomic Scattering Factors" by Peng et al. (1996),
        given by [`read_peng_element_scattering_factor_parameter_table`](https://michael-0brien.github.io/cryojax/api/constants/scattering_factor_parameters/#cryojax.constants.read_peng_element_scattering_factor_parameter_table).

    **Returns:**

    The particular scattering factor parameters stored in
    `scattering_factor_parameter_table` for `atom_types`.
    """  # noqa: E501
    if scattering_factor_parameter_table is None:
        scattering_factor_parameter_table = read_peng_scattering_factor_parameter_table()
    return {
        str(k): np.asarray(v.data[np.asarray(atom_types), ...])
        for k, v in scattering_factor_parameter_table.items()
    }



from abc import abstractmethod
from typing import Generic, TypeVar

import equinox as eqx
from jaxtyping import Array, Complex, Float

from .._image_config import AbstractImageConfig
from .._volume import AbstractVolumeRepresentation


VolRepT = TypeVar("VolRepT", bound="AbstractVolumeRepresentation")


class AbstractMultisliceIntegrator(eqx.Module, Generic[VolRepT], strict=True):
    """Base class for a multislice integration scheme."""

    @abstractmethod
    def integrate(
        self,
        volume_representation: VolRepT,
        image_config: AbstractImageConfig,
        amplitude_contrast_ratio: Float[Array, ""] | float,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]:
        raise NotImplementedError



from .base_integrator import (
    AbstractVolumeIntegrator as AbstractVolumeIntegrator,
)
from .fft_atom_projection import FFTAtomProjection as FFTAtomProjection
from .fourier_voxel_extract import (
    EwaldSphereExtraction as EwaldSphereExtraction,
    FourierSliceExtraction as FourierSliceExtraction,
)
from .gaussian_projection import (
    GaussianMixtureProjection as GaussianMixtureProjection,
)
from .real_voxel_projection import RealVoxelProjection as RealVoxelProjection



from typing import TypeAlias

import numpy as np
from jaxtyping import Array, Bool, Complex, Float, Inexact, Int


NDArrayLike: TypeAlias = Array | np.ndarray
BoolLike: TypeAlias = bool | Bool[Array | np.ndarray, ""]
ComplexLike: TypeAlias = complex | Complex[Array | np.ndarray, ""]
FloatLike: TypeAlias = float | Float[Array | np.ndarray, ""]
InexactLike: TypeAlias = complex | float | Inexact[Array | np.ndarray, ""]
IntLike: TypeAlias = int | Int[Array | np.ndarray, ""]



from time import time

import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
from cryojax.constants import PengScatteringFactorParameters
from cryojax.dataset import RelionParticleParameterFile
from cryojax.io import read_atoms_from_pdb
from cryojax.ndimage import operators as op
from cryojax.rotations import SO3
from jaxtyping import PRNGKeyArray


def setup(num_images, path_to_pdb, path_to_starfile):
    @eqx.filter_vmap(in_axes=(0, None))
    def make_particle_parameters(key: PRNGKeyArray, config: cxs.BasicImageConfig):
        """Generate random parameters."""
        # Pose
        # ... instantiate rotations
        key, subkey = jax.random.split(
            key
        )  # split the key to use for the next random number
        rotation = SO3.sample_uniform(subkey)

        # ... now in-plane translation
        ny, nx = config.shape

        key, subkey = jax.random.split(key)  # do this everytime you use a key!!
        offset_in_angstroms = (
            jax.random.uniform(subkey, (2,), minval=-0.1, maxval=0.1)
            * jnp.asarray((nx, ny))
            / 2
            * config.pixel_size
        )
        # ... build the pose
        pose = cxs.EulerAnglePose.from_rotation_and_translation(
            rotation, offset_in_angstroms
        )

        # CTF Parameters
        # ... defocus
        key, subkey = jax.random.split(key)
        defocus_in_angstroms = jax.random.uniform(subkey, (), minval=10000, maxval=15000)
        # ... astigmatism
        key, subkey = jax.random.split(key)
        astigmatism_in_angstroms = jax.random.uniform(subkey, (), minval=0, maxval=100)
        key, subkey = jax.random.split(key)
        astigmatism_angle = jax.random.uniform(subkey, (), minval=0, maxval=jnp.pi)
        # Now non-random values
        spherical_aberration_in_mm = 2.7
        amplitude_contrast_ratio = 0.1
        # Build the CTF
        transfer_theory = cxs.ContrastTransferTheory(
            ctf=cxs.AstigmaticCTF(
                defocus_in_angstroms=defocus_in_angstroms,
                astigmatism_in_angstroms=astigmatism_in_angstroms,
                astigmatism_angle=astigmatism_angle,
                spherical_aberration_in_mm=spherical_aberration_in_mm,
            ),
            amplitude_contrast_ratio=amplitude_contrast_ratio,
        )

        return {
            "image_config": config,
            "pose": pose,
            "transfer_theory": transfer_theory,
        }

    # Generate particle parameters. First, the image config
    pad_options = dict(shape=(200, 200))

    config = cxs.BasicImageConfig(
        shape=(150, 150),
        pixel_size=2.0,
        voltage_in_kilovolts=300.0,
        pad_options=pad_options,
    )
    # ... RNG keys
    keys = jax.random.split(jax.random.key(0), num_images)
    # ... make parameters
    particle_parameters = make_particle_parameters(keys, config)

    parameter_file = RelionParticleParameterFile(
        path_to_starfile=path_to_starfile,
        mode="w",  # writing mode!
        exists_ok=True,  # in case the file already exists
        broadcasts_image_config=True,
        pad_options=pad_options,
    )
    parameter_file.append(particle_parameters)

    atom_positions, atom_types, atom_properties = read_atoms_from_pdb(
        path_to_pdb,
        center=True,
        loads_properties=True,
        selection_string="name CA",  # C-Alphas for simplicity
    )
    scattering_parameters = PengScatteringFactorParameters(atom_types)
    volume_gmm = cxs.GaussianMixtureVolume.from_tabulated_parameters(
        atom_positions,
        scattering_parameters,
        extra_b_factors=atom_properties["b_factors"],
    )

    real_voxel_grid = volume_gmm.to_real_voxel_grid(shape=(150, 150, 150), voxel_size=1.0)
    volume_fourier_grid = cxs.FourierVoxelGridVolume.from_real_voxel_grid(
        real_voxel_grid, pad_scale=2
    )
    # print(atom_properties["b_factors"].mean() * 8 * jnp.pi**2)
    atom_volume = cxs.IndependentAtomVolume(
        position_pytree=atom_positions,
        scattering_factor_pytree=op.FourierGaussian(amplitude=1.0, b_factor=10.0),
    )
    return (
        particle_parameters,
        parameter_file,
        volume_gmm,
        atom_volume,
        volume_fourier_grid,
    )


# def compute_image(particle_parameters, volume, volume_integrator, rng_key):
#     # Build image model, including normalization within a circular mask
#     # around each particle
#     image_config, pose, transfer_theory = (
#         particle_parameters["image_config"],
#         particle_parameters["pose"],
#         particle_parameters["transfer_theory"],
#     )
#     mask = tf.CircularCosineMask(
#         coordinate_grid=image_config.coordinate_grid_in_angstroms,
#         radius=150.0,
#         rolloff_width=0.0,
#         xy_offset=pose.offset_in_angstroms,
#     )
#     image_model = cxs.make_image_model(
#         volume_parametrisation=volume,
#         pose=pose,
#         image_config=image_config,
#         transfer_theory=transfer_theory,
#         volume_integrator=volume_integrator,
#         normalizes_signal=True,
#         signal_region=mask.get() == 1.0,
#     )
#     # Build noise model at a randomly sampled SNR within a
#     # uniform range, then simulate
#     simulator_rng_key, snr_rng_key = jax.random.split(rng_key, 2)
#     snr = jax.random.uniform(snr_rng_key, minval=0.01, maxval=0.1)
#     noise_model = cxs.UncorrelatedGaussianNoiseModel(
#         image_model,
#         variance=1.0,
#         signal_scale_factor=jnp.sqrt(snr),
#     )

#     return noise_model.sample(rng_key=simulator_rng_key)


# def benchmark_projection_microscope_noise_no_vmap(
#     n_iterations, path_to_pdb, path_to_starfile
# ):
#     _, parameter_file, volume_gmm, volume_fourier_grid = setup(
#         path_to_pdb, path_to_starfile
#     )

#     time_list = []
#     for _ in range(n_iterations + 1):
#         start_time = time()
#         gmm_image = compute_image(
#             parameter_file[0],
#             volume_gmm,
#             cxs.GaussianMixtureProjection(),
#             jax.random.key(1234),
#         )
#         gmm_image.block_until_ready()
#         end_time = time()
#         time_list.append(end_time - start_time)
#     gmm_avg_time = jnp.mean(jnp.array(time_list[1:]))
#     gmm_std_time = jnp.std(jnp.array(time_list[1:]))
#     print(f"GMM: {gmm_avg_time:.4f} +/- {gmm_std_time:.4f} s")

#     time_list = []
#     for _ in range(n_iterations + 1):
#         start_time = time()
#         gmm_image = eqx.filter_jit(compute_image)(
#             parameter_file[0],
#             volume_gmm,
#             cxs.GaussianMixtureProjection(),
#             jax.random.key(1234),
#         )
#         gmm_image.block_until_ready()
#         end_time = time()
#         time_list.append(end_time - start_time)
#     jit_gmm_avg_time = jnp.mean(jnp.array(time_list[1:]))
#     jit_gmm_std_time = jnp.std(jnp.array(time_list[1:]))
#     print(
#         f"JIT, GMM: {jit_gmm_avg_time:.4f} +/- {jit_gmm_std_time:.4f} s"
#     )

#     time_list = []
#     for _ in range(n_iterations + 1):
#         start_time = time()
#         fs_image = compute_image(
#             parameter_file[0],
#             volume_fourier_grid,
#             cxs.FourierSliceExtraction(),
#             jax.random.key(1234),
#         )
#         fs_image.block_until_ready()
#         end_time = time()
#         time_list.append(end_time - start_time)
#     fs_avg_time = jnp.mean(jnp.array(time_list[1:]))
#     fs_std_time = jnp.std(jnp.array(time_list[1:]))
#     print(f"Fourier Slice: {fs_avg_time:.4f} +/- {fs_std_time:.4f} s")

#     time_list = []
#     for _ in range(n_iterations + 1):
#         start_time = time()
#         fs_image = eqx.filter_jit(compute_image)(
#             parameter_file[0],
#             volume_fourier_grid,
#             cxs.FourierSliceExtraction(),
#             jax.random.key(1234),
#         )
#         fs_image.block_until_ready()
#         end_time = time()
#         time_list.append(end_time - start_time)
#     fs_avg_time = jnp.mean(jnp.array(time_list[1:]))
#     fs_std_time = jnp.std(jnp.array(time_list[1:]))
#     print(f"JIT, Fourier Slice: {fs_avg_time:.4f} +/- {fs_std_time:.4f} s")


@eqx.filter_vmap(in_axes=(eqx.if_array(0), eqx.if_array(0), eqx.if_array(0), None, None))
def simulate_image_nojit(
    image_config, pose, transfer_theory, potential, volume_integrator
):
    image_model = cxs.make_image_model(
        volume_parametrization=potential,
        image_config=image_config,
        pose=pose,
        transfer_theory=transfer_theory,
        volume_integrator=volume_integrator,
    )
    return image_model.simulate()


simulate_image_jit = eqx.filter_jit(simulate_image_nojit)


def benchmark_fourier_slice_vs_gmm(
    n_iterations, num_images, path_to_pdb, path_to_starfile
):
    particle_parameters, _, volume_gmm, atom_volume, volume_fourier_grid = setup(
        num_images, path_to_pdb, path_to_starfile
    )

    image_config, pose, transfer_theory = (
        particle_parameters["image_config"],
        particle_parameters["pose"],
        particle_parameters["transfer_theory"],
    )

    time_list = []
    for _ in range(n_iterations + 1):
        start_time = time()
        fft_image = simulate_image_nojit(
            image_config,
            pose,
            transfer_theory,
            atom_volume,
            cxs.FFTAtomProjection(eps=1e-16),
        )
        fft_image.block_until_ready()
        end_time = time()
        time_list.append(end_time - start_time)
    fft_avg_time = jnp.mean(jnp.array(time_list[1:]))
    fft_std_time = jnp.std(jnp.array(time_list[1:]))
    print(f"FFTproj (no JIT): {1000 * fft_avg_time:.2f} +/- {1000 * fft_std_time:.2f} ms")

    time_list = []
    for _ in range(n_iterations + 1):
        start_time = time()
        fft_image = simulate_image_jit(
            image_config,
            pose,
            transfer_theory,
            atom_volume,
            cxs.FFTAtomProjection(eps=1e-16),
        )
        fft_image.block_until_ready()
        end_time = time()
        time_list.append(end_time - start_time)
    fft_avg_time = jnp.mean(jnp.array(time_list[1:]))
    fft_std_time = jnp.std(jnp.array(time_list[1:]))
    print(f"FFTproj (JIT): {1000 * fft_avg_time:.2f} +/- {1000 * fft_std_time:.2f} ms")

    time_list = []
    for _ in range(n_iterations + 1):
        start_time = time()
        gmm_image = simulate_image_nojit(
            image_config,
            pose,
            transfer_theory,
            volume_gmm,
            cxs.GaussianMixtureProjection(),
        )
        gmm_image.block_until_ready()
        end_time = time()
        time_list.append(end_time - start_time)
    gmm_avg_time = jnp.mean(jnp.array(time_list[1:]))
    gmm_std_time = jnp.std(jnp.array(time_list[1:]))
    print(f"GMM (no JIT): {1000 * gmm_avg_time:.2f} +/- {1000 * gmm_std_time:.2f} ms")

    time_list = []
    for _ in range(n_iterations + 1):
        start_time = time()
        gmm_image = simulate_image_jit(
            image_config,
            pose,
            transfer_theory,
            volume_gmm,
            cxs.GaussianMixtureProjection(),
        )
        gmm_image.block_until_ready()
        end_time = time()
        time_list.append(end_time - start_time)
    jit_gmm_avg_time = jnp.mean(jnp.array(time_list[1:]))
    jit_gmm_std_time = jnp.std(jnp.array(time_list[1:]))
    print(
        f"GMM (JIT): {1000 * jit_gmm_avg_time:.2f} +/- {1000 * jit_gmm_std_time:.2f} ms"
    )

    time_list = []
    for _ in range(n_iterations + 1):
        start_time = time()
        fs_image = simulate_image_nojit(
            image_config,
            pose,
            transfer_theory,
            volume_fourier_grid,
            cxs.FourierSliceExtraction(),
        )
        fs_image.block_until_ready()
        end_time = time()
        time_list.append(end_time - start_time)
    fs_avg_time = jnp.mean(jnp.array(time_list[1:]))
    fs_std_time = jnp.std(jnp.array(time_list[1:]))
    print(
        f"Fourier Slice (no JIT): {1000 * fs_avg_time:.2f} "
        f"+/- {1000 * fs_std_time:.2f} ms"
    )

    time_list = []
    for _ in range(n_iterations + 1):
        start_time = time()
        fs_image = simulate_image_jit(
            image_config,
            pose,
            transfer_theory,
            volume_fourier_grid,
            cxs.FourierSliceExtraction(),
        )
        fs_image.block_until_ready()
        end_time = time()
        time_list.append(end_time - start_time)
    fs_avg_time = jnp.mean(jnp.array(time_list[1:]))
    fs_std_time = jnp.std(jnp.array(time_list[1:]))
    print(
        f"Fourier Slice (JIT): {1000 * fs_avg_time:.2f} +/- {1000 * fs_std_time:.2f} ms"
    )


if __name__ == "__main__":
    n_iterations, n_images = 10, 100
    print(
        f"Benchmarking image simulation of {n_images} images "
        f"averaged over {n_iterations} iterations"
    )
    path_to_pdb = "../data/thyroglobulin_initial.pdb"
    path_to_starfile = "../outputs/particles.star"
    benchmark_fourier_slice_vs_gmm(n_iterations, n_images, path_to_pdb, path_to_starfile)



from typing import Any, ClassVar, Literal
from typing_extensions import override

import jax
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float

from ...coordinates import make_frequency_grid
from ...ndimage import (
    block_reduce_downsample,
    convert_fftn_to_rfftn,
    irfftn,
    operators as op,
    resize_with_crop_or_pad,
    rfftn,
)
from .._image_config import AbstractImageConfig
from .._volume import IndependentAtomVolume
from .base_integrator import AbstractVolumeIntegrator


try:
    import jax_finufft as jnufft

    JAX_FINUFFT_IMPORT_ERROR = None
except ModuleNotFoundError as err:
    jnufft = None
    JAX_FINUFFT_IMPORT_ERROR = err


class FFTAtomProjection(
    AbstractVolumeIntegrator[IndependentAtomVolume],
    strict=True,
):
    """Integrate atomic parametrization of a volume onto
    the exit plane using non-uniform FFTs plus convolution.
    """

    sampling_mode: Literal["average", "point"]
    upsample_factor: int | None
    shape: tuple[int, int] | None
    eps: float
    opts: Any

    is_projection_approximation: ClassVar[bool] = True

    def __init__(
        self,
        *,
        sampling_mode: Literal["average", "point"] = "average",
        upsample_factor: int | None = None,
        shape: tuple[int, int] | None = None,
        eps: float = 1e-6,
        opts: Any = None,
    ):
        """**Arguments:**

        - `sampling_mode`:
            If `'average'`, convolve with a box function to sample the
            projected volume at a pixel to be the average value of the
            underlying continuous function. If `'point'`, the volume at
            a pixel will be point sampled.
        - `upsample_factor`:
            If provided, first compute an upsampled version of the
            image at pixel size `image_config.pixel_size / upsample_factor`.
            Then, downsample with `cryojax.ndimage.block_reduce_downsample`
            to locally average to the correct pixel size. This is useful
            for reducing aliasing.
        - `shape`:
            If given, first compute the image at `shape`, then
            pad or crop to `image_config.padded_shape`.
        - `eps`:
            See [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
            for documentation.
        - `opts`:
            A `jax_finufft.options.Opts` or `jax_finufft.options.NestedOpts`
            dataclass.
            See [`jax-finufft`](https://github.com/flatironinstitute/jax-finufft)
            for documentation.
        """
        if jnufft is None:
            raise RuntimeError(
                "Tried to use the `FFTAtomProjection` "
                "class, but `jax-finufft` is not installed. "
                "See https://github.com/flatironinstitute/jax-finufft "
                "for installation instructions."
            ) from JAX_FINUFFT_IMPORT_ERROR
        if sampling_mode not in ["average", "point"]:
            raise ValueError(
                "`sampling_mode` in `FFTAtomProjection` "
                "must be either 'average' for averaging within a "
                "pixel or 'point' for point sampling. Got "
                f"`sampling_mode = {sampling_mode}`."
            )
        self.sampling_mode = sampling_mode
        self.upsample_factor = upsample_factor
        self.shape = shape
        self.eps = eps
        self.opts = opts

    def __check_init__(self):
        if self.upsample_factor is not None:
            if self.upsample_factor % 2 == 0:
                raise ValueError(
                    f"Set `upsample_factor = {self.upsample_factor}` when instantiating "
                    "`FFTAtomProjection`, but only odd `upsample_factor` are supported."
                )

    @override
    def integrate(
        self,
        volume_representation: IndependentAtomVolume,
        image_config: AbstractImageConfig,
        outputs_real_space: bool = False,
    ) -> (
        Complex[
            Array,
            "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}",
        ]
        | Float[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]
    ):
        """Compute a projection from scattering factors per atom type
        from the `IndependentAtomVolume`.

        **Arguments:**

        - `volume_representation`:
            The volume representation.
        - `image_config`:
            The configuration of the resulting image.
        - `outputs_real_space`:
            If `True`, return the image in real space. Otherwise,
            return in fourier.

        **Returns:**

        The integrated volume in real or fourier space at the
        `AbstractImageConfig.padded_shape`.
        """  # noqa: E501
        u = self.upsample_factor
        pixel_size = image_config.pixel_size
        shape = image_config.padded_shape if self.shape is None else self.shape
        if u is None:
            shape_u, pixel_size_u = shape, pixel_size
        else:
            shape_u, pixel_size_u = (u * shape[0], u * shape[1]), pixel_size / u
        if shape_u == image_config.padded_shape:
            frequency_grid = image_config.padded_full_frequency_grid_in_angstroms
        else:
            frequency_grid = make_frequency_grid(
                shape_u, pixel_size_u, outputs_rfftfreqs=False
            )
        frequency_grid = jnp.fft.fftshift(frequency_grid, axes=(0, 1))
        proj_kernel = lambda pos, kernel: _project_with_nufft(
            shape_u,
            pixel_size_u,
            pos,
            kernel,
            frequency_grid,
            eps=self.eps,
            opts=self.opts,
        )
        # Compute projection over atom types
        fourier_projection = jax.tree.reduce(
            lambda x, y: x + y,
            jax.tree.map(
                proj_kernel,
                volume_representation.position_pytree,
                volume_representation.scattering_factor_pytree,
                is_leaf=lambda x: isinstance(x, op.AbstractFourierOperator),
            ),
        )
        # Apply anti-aliasing filter
        if self.sampling_mode == "average":
            antialias_fn = op.FourierSinc(box_width=pixel_size_u)
            fourier_projection *= antialias_fn(frequency_grid)
        # Shift zero frequency component to corner and convert to
        # rfft
        fourier_projection = convert_fftn_to_rfftn(
            jnp.fft.ifftshift(fourier_projection), mode="real"
        )
        if self.shape is None:
            if u is None:
                return (
                    irfftn(fourier_projection, s=shape)
                    if outputs_real_space
                    else fourier_projection
                )
            else:
                projection = _block_average(irfftn(fourier_projection, s=shape_u), u)
                return projection if outputs_real_space else rfftn(projection)
        else:
            projection = irfftn(fourier_projection, s=shape_u)
            if u is not None:
                projection = _block_average(projection, u)
            projection = resize_with_crop_or_pad(projection, image_config.padded_shape)
            return projection if outputs_real_space else rfftn(projection)


def _project_with_nufft(shape, ps, pos, kernel, freqs, eps=1e-6, opts=None):
    assert jnufft is not None
    # Get x and y coordinates
    positions_xy = pos[:, :2]
    # Normalize coordinates betweeen -pi and pi
    ny, nx = shape
    box_xy = ps * jnp.asarray((nx, ny))
    positions_periodic = 2 * jnp.pi * positions_xy / box_xy
    # Unpack
    x, y = positions_periodic[:, 0], positions_periodic[:, 1]
    n_atoms = x.size
    area_element = ps**2
    # Compute
    fourier_projection = kernel(freqs) * (
        jnufft.nufft1(
            shape,
            jnp.full((n_atoms,), 1.0 + 0.0j),
            y,
            x,
            eps=eps,
            opts=opts,
            iflag=-1,
        )
        / area_element
    )

    return fourier_projection


def _block_average(x, factor):
    return block_reduce_downsample(x, factor, jax.lax.add) / factor**x.ndim



import warnings as _warnings
from typing import Any as _Any

from . import operators as operators, transforms as transforms
from ._downsample import (
    block_reduce_downsample as block_reduce_downsample,
    fourier_crop_downsample as fourier_crop_downsample,
    fourier_crop_downsample_to_shape as fourier_crop_downsample_to_shape,
)
from ._edges import (
    crop_to_shape as crop_to_shape,
    crop_to_shape_with_center as crop_to_shape_with_center,
    pad_to_shape as pad_to_shape,
    resize_with_crop_or_pad as resize_with_crop_or_pad,
)
from ._fft import (
    fftn as fftn,
    ifftn as ifftn,
    irfftn as irfftn,
    rfftn as rfftn,
)
from ._fourier_statistics import (
    compute_binned_powerspectrum as compute_binned_powerspectrum,
    compute_fourier_ring_correlation as compute_fourier_ring_correlation,
    compute_fourier_shell_correlation as compute_fourier_shell_correlation,
)
from ._fourier_utils import (
    convert_fftn_to_rfftn as convert_fftn_to_rfftn,
    enforce_self_conjugate_rfftn_components as enforce_self_conjugate_rfftn_components,
)
from ._map_coordinates import (
    compute_spline_coefficients as compute_spline_coefficients,
    map_coordinates as map_coordinates,
    map_coordinates_spline as map_coordinates_spline,
)
from ._normalize import normalize_image as normalize_image, rescale_image as rescale_image
from ._radial_average import (
    compute_binned_radial_average as compute_binned_radial_average,
    interpolate_radial_average_on_grid as interpolate_radial_average_on_grid,
)
from ._rescale_pixel_size import (
    rescale_pixel_size as rescale_pixel_size,
)


def __getattr__(name: str) -> _Any:
    # Future deprecations
    if name == "downsample_with_fourier_cropping":
        _warnings.warn(
            "'downsample_with_fourier_cropping' is deprecated"
            "has been renamed to 'fourier_crop_downsample'. "
            "The old name will be deprecated in cryoJAX 0.6.0.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return fourier_crop_downsample
    if name == "downsample_to_shape_with_fourier_cropping":
        _warnings.warn(
            "'downsample_to_shape_with_fourier_cropping' is deprecated"
            "has been renamed to 'fourier_crop_downsample_to_shape'. "
            "The old name will be deprecated in cryoJAX 0.6.0.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return fourier_crop_downsample_to_shape

    raise AttributeError(f"cannot import name '{name}' from 'cryojax.ndimage'.")
